import uuid
from datetime import datetime
from urllib.parse import parse_qs, quote_plus, unquote_plus, urlparse

import requests
from bs4 import BeautifulSoup
from slugify import slugify as _dash_slugify

from .settings import BASE_URL, S3_URL_PREFIX, S3_URL_PREFIX_ENCODED, SIGNED_URL_PREFIX


class InvalidNotionIdentifier(Exception):
    pass


def now():
    return int(datetime.now().timestamp() * 1000)


def extract_id(url_or_id):
    """
    Extract the block/page ID from a Notion.so URL -- if it's a bare page URL, it will be the
    ID of the page. If there's a hash with a block ID in it (from clicking "Copy Link") on a
    block in a page), it will instead be the ID of that block. If it's already in ID format,
    it will be passed right through.
    """
    input_value = url_or_id
    if url_or_id.startswith(BASE_URL):
        url_or_id = (
            url_or_id.split("#")[-1]
            .split("/")[-1]
            .split("&p=")[-1]
            .split("?")[0]
            .split("-")[-1]
        )
    try:
        return str(uuid.UUID(url_or_id))
    except ValueError:
        raise InvalidNotionIdentifier(input_value)


def get_embed_data(source_url):

    return requests.get(
        "https://api.embed.ly/1/oembed?key=421626497c5d4fc2ae6b075189d602a2&url={}".format(
            source_url
        )
    ).json()


def get_embed_link(source_url):

    data = get_embed_data(source_url)

    if "html" not in data:
        return source_url

    url = list(BeautifulSoup(data["html"], "html.parser").children)[0]["src"]

    return parse_qs(urlparse(url).query)["src"][0]


def add_signed_prefix_as_needed(url, client=None, id=""):

    if url is None:
        return

    if url.startswith(S3_URL_PREFIX):
        url = SIGNED_URL_PREFIX + quote_plus(url) + "?table=block&id=" + id
        if client:
            url = client.session.head(url).headers.get("Location")

    return url


def remove_signed_prefix_as_needed(url):
    if url is None:
        return
    if url.startswith(SIGNED_URL_PREFIX):
        return unquote_plus(url[len(S3_URL_PREFIX) :])
    elif url.startswith(S3_URL_PREFIX_ENCODED):
        parsed = urlparse(url.replace(S3_URL_PREFIX_ENCODED, S3_URL_PREFIX))
        return f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
    else:
        return url


def slugify(original):
    return _dash_slugify(original).replace("-", "_")


def get_by_path(path, obj, default=None):

    if isinstance(path, str):
        path = path.split(".")

    value = obj

    # try to traverse down the sequence of keys defined in the path, to get the target value if it exists
    try:
        for key in path:
            if isinstance(value, list):
                key = int(key)
            value = value[key]
    except (KeyError, TypeError, IndexError):
        value = default

    return value



from copy import deepcopy

from .logger import logger
from .operations import build_operation
from .utils import extract_id, get_by_path


class Record:

    # if a subclass has a list of ids that should be update when child records are removed, it should specify the key here
    child_list_key = None

    def __init__(self, client, id, *args, **kwargs):
        self._client = client
        self._id = extract_id(id)
        self._callbacks = []
        if self._client._monitor is not None:
            self._client._monitor.subscribe(self)

    @property
    def id(self):
        return self._id

    @property
    def role(self):
        return self._client._store.get_role(self._table, self.id)

    def _str_fields(self):
        """
        Determines the list of fields to include in the __str__ representation. Override and extend this in subclasses.
        """
        return ["id"]

    def __str__(self):
        return ", ".join(
            [
                f"{field}={repr(getattr(self, field))}"
                for field in self._str_fields()
                if getattr(self, field, "")
            ]
        )

    def __repr__(self):
        return f"<{self.__class__.__name__} ({self})>"

    def refresh(self):
        """
        Update the cached data for this record from the server (data for other records may be updated as a side effect).
        """
        self._get_record_data(force_refresh=True)

    def _convert_diff_to_changelist(self, difference, old_val, new_val):
        changed_values = set()
        for operation, path, values in deepcopy(difference):
            path = path.split(".") if isinstance(path, str) else path
            if operation in ["add", "remove"]:
                path.append(values[0][0])
            while isinstance(path[-1], int):
                path.pop()
            changed_values.add(".".join(map(str, path)))
        return [
            (
                "changed_value",
                path,
                (get_by_path(path, old_val), get_by_path(path, new_val)),
            )
            for path in changed_values
        ]

    def add_callback(self, callback, callback_id=None, extra_kwargs={}):
        assert callable(
            callback
        ), "The callback must be a 'callable' object, such as a function."
        callback_obj = self._client._store.add_callback(
            self, callback, callback_id=callback_id, extra_kwargs=extra_kwargs
        )
        self._callbacks.append(callback_obj)
        return callback_obj

    def remove_callbacks(self, callback_or_callback_id_prefix=None):
        if callback_or_callback_id_prefix is None:
            for callback_obj in list(self._callbacks):
                self._client._store.remove_callbacks(
                    self._table, self.id, callback_or_callback_id_prefix=callback_obj
                )
            self._callbacks = []
        else:
            self._client._store.remove_callbacks(
                self._table,
                self.id,
                callback_or_callback_id_prefix=callback_or_callback_id_prefix,
            )
            if callback_or_callback_id_prefix in self._callbacks:
                self._callbacks.remove(callback_or_callback_id_prefix)

    def _get_record_data(self, force_refresh=False):
        return self._client.get_record_data(
            self._table, self.id, force_refresh=force_refresh
        )

    def get(self, path=[], default=None, force_refresh=False):
        """
        Retrieve cached data for this record. The `path` is a list (or dot-delimited string) the specifies the field
        to retrieve the value for. If no path is supplied, return the entire cached data structure for this record.
        If `force_refresh` is set to True, we force_refresh the data cache from the server before reading the values.
        """
        return get_by_path(
            path, self._get_record_data(force_refresh=force_refresh), default=default
        )

    def set(self, path, value):
        """
        Set a specific `value` (under the specific `path`) on the record's data structure on the server.
        """
        self._client.submit_transaction(
            build_operation(id=self.id, path=path, args=value, table=self._table)
        )

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return hash(self.id)



import logging
import os

from .settings import LOG_FILE

NOTIONPY_LOG_LEVEL = os.environ.get("NOTIONPY_LOG_LEVEL", "warning").lower()

logger = logging.getLogger("notion")


def enable_debugging():
    set_log_level(logging.DEBUG)


def set_log_level(level):
    logger.setLevel(level)
    handler.setLevel(level)


# if NOTIONPY_LOG_LEVEL == "disabled":
#     handler = logging.NullHandler()
#     logger.addHandler(handler)
# else:
#     handler = logging.FileHandler(LOG_FILE)
#     formatter = logging.Formatter("\n%(asctime)s - %(levelname)s - %(message)s")
#     handler.setFormatter(formatter)
#     logger.addHandler(handler)

#     if NOTIONPY_LOG_LEVEL == "debug":
#         set_log_level(logging.DEBUG)
#     elif NOTIONPY_LOG_LEVEL == "info":
#         set_log_level(logging.INFO)
#     elif NOTIONPY_LOG_LEVEL == "warning":
#         set_log_level(logging.WARNING)
#     elif NOTIONPY_LOG_LEVEL == "error":
#         set_log_level(logging.ERROR)
#     else:
#         raise Exception("Invalid value for environment variable NOTIONPY_LOG_LEVEL: {}".format(NOTIONPY_LOG_LEVEL))

handler = logging.NullHandler()
logger.addHandler(handler)
